#!/usr/bin/perl
use strict;

my ($module, $baseDir, $compareToDir) = @ARGV[0..2];
my $exceptionsRegExp='^(\..*|tools|plugin|galleries)$';
my @newFiles;
my @filesToDiff;
my @filesToEdit;
my $tempFile;
my $fileModifications;

cdCompareToDir($compareToDir);

foreach (@filesToDiff) {
  $tempFile = compareDir2baseDir($_);
  diffFiles($tempFile, $_);
}

# Printing files to edit
print "## Files To Edit:\n";
foreach my $fileToEdit (@filesToEdit) {
  $_ = $fileToEdit;
  s/$baseDir\/*//;
  print "##   $_\n";
}

print "\n";

# Printing new files
if (scalar(@newFiles)) {
  print "## Included Files:\n";
  foreach my $newFile (@newFiles) {
    $_ = $newFile;
    s/$compareToDir\/*//;
    print "##   $_\n";
  }
  print "\n";
}

# Printing modifications
print "##############################################################\n";
print $fileModifications;


sub cdCompareToDir {
  my ($path) = @_;

  $_ = qx/basename "$path"/;
  chomp;
  opendir(DIR, $path) or die "$path: $!";
  my @files = sort grep { not /$exceptionsRegExp\z/ } readdir DIR;
  foreach my $entry (@files) {
    $_ = $path . '/' . $entry;
    if (-d) {
      cdCompareToDir($_);
    }
    else {
      $tempFile = compareDir2baseDir($_);
      if (-f $tempFile) {
        push @filesToDiff, $_;
      }
      else {
        push @newFiles, $_;
      }
    }
  }
}

sub compareDir2baseDir {
  my ($dirName) = @_;

  $dirName =~ s/$compareToDir/$baseDir/;

  $dirName;
}

sub diffFiles {
  my ($baseFile, $compareToFile) = @_;

  my $diff = `diff -c "$baseFile" "$compareToFile"` or return;

  my @modifications = split /^[*]{15}\n/m, $diff;

  my (@findAfter, @findBefore, @new, @type);

  shift @modifications;
  my $temp;
  my ($replaceWhat, $replaceWith, $findAfter, $findBefore, $new, $type);
  while (my $elt = shift @modifications) {
    $elt =~ s/^\*{3}\s\d+,\d+\s\*{4}\n// or next;
    
    my $cpy_elt = $elt;

    # Looking for replacements
    if ($cpy_elt !~ /^-{3}\s\d+,\d+\s-{4}\n/) {{
      # Replace mode
      while ($cpy_elt =~ s/^[^!].*\n//) {};
      $replaceWhat = '';
      $replaceWhat .= " $1\n" while ($cpy_elt =~ s/^[!](.*)\n//);
      while ($cpy_elt =~ s/^[^-].*\n//) {};
      $cpy_elt =~ s/-{3}\s\d+,\d+\s-{4}// or last;
      while ($cpy_elt =~ s/^[^!].*\n//) {};
      $replaceWith = '';
      $replaceWith .= " $1\n" while ($cpy_elt =~ s/^[!](.*)\n//);

      $findBefore = '';
      $findBefore .= "$1\n" while ($cpy_elt =~ s/^([^-+!].*)\n//);

      $replaceWhat = reArrange($replaceWhat);
      $replaceWith = reArrange($replaceWith);
      $findBefore = reArrange($findBefore);


      # We check that we're not actually replacing spaces with other spaces
      # which is (I hope you agree) useless
      unless ($replaceWhat =~ /^(\s|\n)*$/ and $replaceWith =~ /^(\s|\n)*$/) {
        last if ($replaceWhat eq $replaceWith);
        push @findAfter, $replaceWhat;
        push @findBefore, $findBefore;
        push @new, $replaceWith;
        push @type, 'REPLACE';
      }
    }}

    # Insert/Remove mode
    while ($elt =~ s/^[^-].*\n//) {};
    $elt =~ s/^-{3}\s\d+,\d+\s-{4}\n//;

    $findAfter = '';
    $findAfter .= " $1\n" while ($elt =~ s/^[^+-](.*)\n//);

    $new = '';
    while ($elt =~ s/^([+-])(.*)\n//) {
      $new .= " $2\n";
      if ($1 eq '+') {
        $type = 'ADD';
      }
      else {
        $type = 'REMOVE';
      }
    }

    $findBefore = '';
    $findBefore .= "$1\n" while ($elt =~ s/^(\s+.*)\n//);

    $findAfter = reArrange($findAfter);
    $findBefore = reArrange($findBefore);
    $new = reArrange($new);

    unless ($new eq '') {
      push @findBefore, $findBefore;
      push @findAfter, $findAfter;
      push @new, $new;
      push @type, $type;
    }
  }

  # If the differences detected by diff were rejected, we go to next element

  $fileModifications .=  "#\n#-----[ OPEN ]------------------------------------------\n#\n";
  $_ = $baseFile;
  s/$baseDir\/*//;
  $fileModifications .= "$_\n";
  push @filesToEdit, $_;

  while (my $type = shift @type) {
    $fileModifications .= "#-----[ FIND ]------------------------------------------\n#\n";
    if ($findAfter[0] !~ /^\s*$/s) {
      $fileModifications .= $findAfter[0];
      $temp = $type;
      $temp = 'AFTER, ' . $temp unless ($type eq 'REPLACE');
    }
    else {
      $fileModifications .= $findBefore[0];
      if ($type eq 'REPLACE') {
        # We don't want to actually replace $findBefore[0], we just show it in order
        # for the user to be able to find the lines to modify
        $type = 'ADD';
      }
      $temp = $type;
      $temp = 'BEFORE, ' . $temp;
    }
    $fileModifications .= "#\n#-----[ $temp ]------------------------------------------\n#\n";
    $fileModifications .= $new[0];

    shift @findAfter;
    shift @findBefore;
    shift @new;
  }

  $fileModifications .= "\n\n";
}

sub reArrange {
  my ($data) = @_;

  my $i = 1;
  my @temp;
  my $temp;

  $data =~ s/^\s*$//mg;
  # Replace tabs with 8 spaces. Dirty but should do the trick most of the times.
  $data =~ s/\t/        /g;
  $data =~ s/^(\n)*((.*\S.*\n)((.*\n)*(.*\S.*\n))?)(\n)*/$2/g;

  return $data if ($data =~ /^[\s\n]*$/);

  $temp = $data;
  $temp =~ s/\n+/\n/g;

  $_ = $temp;
  s/[^\n]//g;
  $_ = length;

  while (@temp = $temp =~ /^\s/mg) {
    if (scalar(@temp) == $_) {
      $data =~ s/^ //mg;
      $temp =~ s/^ //mg;
    }
    else {
      last;
    }
  }

  return $data;
}
